<?php
/**
 * @file
 * Implements DrupalStreamWrapperInterface.
 */

class GCSStreamWrapper implements DrupalStreamWrapperInterface
{

    /**
     * Google Cloud Store connection object.
     */
    protected $gcs = NULL;

    /**
     * Google Cloud Store client object.
     */
    protected $client = NULL;

    /**
     * Google Cloud Store credential object.
     */
    protected $cred = NULL;

    /**
     * Google Cloud Store API domain.
     */
    protected $domain = GCS_API_DOMAIN;

    /**
     * String Instance URI referenced as "gs://bucket/".
     */
    protected $uri;

    /**
     * String Google Cloud Storage client ID.
     */
    protected $clientId;

    /**
     * String Google Cloud Storage service account name.
     */
    protected $clientEmail;

    /**
     * String Google Cloud Store bucket name.
     */
    protected $bucket;

    /**
     * String Google Cloud Store folder name.
     */
    protected $folder;

    /**
     * String Google Cloud Store allowed key.
     */
    protected $key;

    /**
     * String Google Cloud Store allowed keyFilePath.
     */
    protected $keyFilePath;

    /**
     * String Google Cloud Store allowed private_key.
     */
    protected $private_key;

    /**
     * Int A buffer for reading/wrting.
     */
    protected $object = NULL;

    /**
     * Int current read/write position.
     */
    protected $position = 0;

    /**
     * Int total size of the object.
     */
    protected $objectSize = 0;

    /**
     * String buffer.
     */
    protected $buffer = NULL;

    /**
     * Boolean Whether $buffer is active as a write buffer.
     */
    protected $bufferWrite = false;

    /**
     * Int Buffer length.
     */
    protected $bufferLength = 0;

    /**
     * Bool HTTPS flag.
     */
    protected $https = false;

    /**
     * Object constructor.
     */
    public function __construct()
    {
        if (isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] !== false)
        {
            $this->https = true;
        }

        $this->clientId = variable_get('gcs_client_id');
        $this->clientEmail = variable_get('gcs_client_email');
        $this->bucket = variable_get('gcs_bucket_name');
        $this->folder = variable_get('gcs_folder');
        $this->private_key = variable_get('gcs_private_key_id');
        $this->key = variable_get('gcs_public_key');
        $this->keyFilePath = variable_get('gcs_public_key_file_path');
        $this->getGCS();

    }

    /**
     * Sets the stream resource URI.
     *
     * URIs are formatted as "gs://bucket/key"
     *
     * @param string $uri
     *   The URI of the file.
     */
    public function setUri($uri)
    {
        $this->uri = $uri;
    }

    /**
     * Returns the stream resource URI.
     *
     * URIs are formatted as "gs://bucket/key"
     *
     * @return string
     *   Returns the current URI of the instance.
     */
    public function getUri()
    {
        return $this->uri;
    }

    /**
     * Returns a web accessible URL for the resource.
     *
     * @return string
     *   Returns a string containing a web accessible URL for the resource.
     */
    public function getExternalUrl()
    {
        $this->callConstructor();
        $path = explode('/', $this->getLocalPath());

        if ($path[0] === 'styles')
        {
            $path_array_last_index = count($path) - 1;
            if ($path[$path_array_last_index])
            {
                try
                {
                    $url = $this->generateSignedURL($path[$path_array_last_index]);
                }
                catch(Exception $e)
                {
                    watchdog_exception('gcs', $e, 'GCSStreamWrapper::getExternalUrl');
                }

                return $url;
            }
        }
        $url = '';
        $local_path = $this->getLocalPath();
        if ($local_path)
        {
            try
            {
                $url = $this->generateSignedURL($local_path);
            }
            catch(Exception $e)
            {
                watchdog_exception('gcs', $e, 'GCSStreamWrapper::getExternalUrl');
            }
        }
        return $url;
    }

    /**
     * Determine a file's media type.
     *
     * Uses Drupal's mimetype mappings. Returns 'application/octet-stream' if
     * no match is found.
     *
     * @return string
     *   Returns a string representing the file's MIME type
     */
    public static function getMimeType($uri, $mapping = NULL)
    {
        // Load the default file map.
        if (!isset($mapping))
        {
            include_once DRUPAL_ROOT . '/includes/file.mimetypes.inc';
            $mapping = file_mimetype_mapping();
        }

        $extension = '';
        $file_parts = explode('.', basename($uri));

        // Remove the first part: a full filename should not match an extension.
        array_shift($file_parts);
        while ($additional_part = array_pop($file_parts))
        {
            $extension = strtolower($additional_part . ($extension ? '.' . $extension : ''));
            if (isset($mapping['extensions'][$extension]))
            {
                return $mapping['mimetypes'][$mapping['extensions'][$extension]];
            }
        }

        return 'application/octet-stream';
    }

    /**
     * Changes permissions of the resource.
     *
     * This doesn't do anything for the moment so just returns TRUE;
     *
     * @param int $mode
     *   Integer value for the permissions. Consult PHP chmod() documentation
     *   for more information.
     *
     * @return bool
     *   Returns TRUE.
     */
    public function chmod($mode)
    {
        return true;
    }

    /**
     * Returns canonical, absolute path of the resource.
     *
     * @return bool
     *   Returns FALSE as this wrapper does not provide an implementation.
     */
    public function realpath()
    {
        return $this->getLocalPath();
    }

    /**
     * Support for fopen(), file_get_contents(), file_put_contents() etc.
     *
     * @param string $uri
     *   A string containing the URI to the file to open.
     * @param string $mode
     *   The file mode ("r", "wb" etc.).
     * @param int $options
     *   A bit mask of STREAM_USE_PATH and STREAM_REPORT_ERRORS.
     * @param string $opened_path
     *   A string containing the path actually opened.
     *
     * @return bool
     *   Returns TRUE if file was opened successfully.
     *
     * @see http://php.net/manual/en/streamwrapper.stream-open.php
     */
    public function stream_open($uri, $mode, $options, &$opened_path)
    {
        $this->callConstructor();
        $this->uri = $uri;
        // If this stream is being opened for writing, clear the object buffer.
        // Return true as we'll create the object on fflush call.
        if (strpbrk($mode, 'wax'))
        {
            $this->clearBuffer();
            $this->bufferWrite = true;
            return true;
        }

        $target = $this->getTarget($this->uri);
        if ($target)
        {
            try
            {
                $response = $this
                    ->gcs
                    ->objects
                    ->get($this->bucket, $this->folder . '/' . $target);
                if ($response)
                {
                    $this->clearBuffer();
                    $this->bufferWrite = false;
                    $this->objectSize = $response['size'];
                    return true;
                }
            }
            catch(Exception $e)
            {
                watchdog_exception('gcs', $e, 'GCSStreamWrapper::stream_open');
            }
        }

        return false;
    }

    /**
     * Support for fclose().
     *
     * Clears the object buffer and returns TRUE
     *
     * @return bool
     *   TRUE
     *
     * @see http://php.net/manual/en/streamwrapper.stream-close.php
     */
    public function stream_close()
    {
        $this->clearBuffer();
        return true;
    }

    /**
     * Support for flock().
     *
     * @param string $operation
     *   One of the following:
     *   - LOCK_SH to acquire a shared lock (reader).
     *   - LOCK_EX to acquire an exclusive lock (writer).
     *   - LOCK_UN to release a lock (shared or exclusive).
     *   - LOCK_NB if you don't want flock() to block while locking (not
     *     supported on Windows).
     *
     * @return bool
     *   returns TRUE if lock was successful
     *
     * @see http://php.net/manual/en/streamwrapper.stream-lock.php
     */
    public function stream_lock($operation)
    {
        return false;
    }

    /**
     * Support for fread(), file_get_contents() etc.
     *
     * @param int $count
     *   Maximum number of bytes to be read.
     *
     * @return string
     *   The string that was read, or FALSE in case of an error.
     *
     * @see http://php.net/manual/en/streamwrapper.stream-read.php
     */
    public function stream_read($count)
    {

        // If the data is empty, get the data from the object.
        if (!$this->object)
        {
            // If the object is empty, get the object from rackspace.
            $target = $this->getTarget($this->uri);

            if ($target)
            {
                try
                {
                    $service = $this
                        ->gcs
                        ->objects
                        ->get($this->bucket, $this->folder . '/' . $target);
                    if ($service)
                    {
                        $download_url = 'https://' . $this->domain . '/' . $this->bucket . '/' . $this->folder . '/' . $target;
                        $signed_request = $this->client->authorize();
                        $request = new GuzzleHttp\Psr7\Request('GET', $object['mediaLink']);
                        $response = $signed_request->send($request);
                        $reasonPhrase = $response->getReasonPhrase();
                        $code = $response->getStatusCode();
                        if ($response)
                        {
                            $this->object = $response->withStatus($code, $reasonPhrase);
                        }
                    }
                }
                catch(Exception $e)
                {
                    watchdog_exception('gooogle_cdn', $e, 'GCSStreamWrapper::stream_read');
                }
            }
        }

        $data = substr($this->object, $this->position, $count);
        $this->position += $count;
        return $data;
    }

    /**
     * Support for fwrite(), file_put_contents() etc.
     *
     * @param string $data
     *   The string to be written.
     *
     * @return int
     *   The number of bytes written.
     *
     * @see http://php.net/manual/en/streamwrapper.stream-write.php
     */
    public function stream_write($data)
    {

        $data_length = strlen($data);
        $this->buffer .= $data;
        $this->bufferLength += $data_length;
        $this->position += $data_length;

        return $data_length;
    }

    /**
     * Support for feof().
     *
     * @return bool
     *   TRUE if end-of-file has been reached.
     *
     * @see http://php.net/manual/en/streamwrapper.stream-eof.php
     */
    public function stream_eof()
    {
        if (!$this->uri)
        {
            return true;
        }
        return ($this->position >= $this->objectSize);
    }

    /**
     * Support for fseek().
     *
     * @param int $offset
     *   The byte offset to got to.
     * @param string $whence
     *   SEEK_SET, SEEK_CUR, or SEEK_END.
     *
     * @return bool
     *   TRUE on success.
     *
     * @see http://php.net/manual/en/streamwrapper.stream-seek.php
     */
    public function stream_seek($offset, $whence)
    {
        switch ($whence)
        {
            case SEEK_CUR:
                // Set position to current location plus $offset.
                $new_position = $this->position + $offset;
            break;

            case SEEK_END:
                // Set position to eof plus $offset.
                $new_position = $this->objectSize + $offset;
            break;

            case SEEK_SET:
            default:
                // Set position equal to $offset.
                $new_position = $offset;
            break;
        }

        $ret = ($new_position >= 0 && $new_position <= $this->objectSize);
        if ($ret)
        {
            $this->position = $new_position;
        }
        return $ret;
    }

    /**
     * Support for fflush(). Flush current cached stream data to storage.
     *
     * @return bool
     *   TRUE if data was successfully stored (or there was no data to store).
     *
     * @see http://php.net/manual/en/streamwrapper.stream-flush.php
     */
    public function stream_flush()
    {

        if ($this->bufferWrite)
        {
            try
            {

                // $storageService = new Google_Service_Storage( $client );
                $file = new Google_Service_Storage_StorageObject();
                $local_path = $this->folder . '/' . $this->getLocalPath($this->uri);
                $file->setName($local_path);
                $post_body = array(
                    'name' => $local_path,
                    'data' => $this->buffer,
                    'uploadType' => 'media',
                    'mimeType' => GCSStreamWrapper::getMimeType($this->uri) ,
                    'predefinedAcl' => 'publicRead'
                );
                $chunk_size_bytes = 1 * 1024 * 1024;
                $request = $this
                    ->gcs
                    ->objects
                    ->insert($this->bucket, $file, $post_body);
                $object = $this
                    ->gcs
                    ->objects
                    ->get($this->bucket, $local_path);
                $signed_request = $this
                    ->client
                    ->authorize();
                $request = new GuzzleHttp\Psr7\Request('GET', $object['mediaLink']);
                $response = $signed_request->send($request);
                $code = $response->getStatusCode();
                if ($code == '200')
                {
                    return true;
                }
            }
            catch(Exception $e)
            {
                watchdog_exception('gcs', $e, 'GCSStreamWrapper::stream_flush');
            }
        }
        $this->clearBuffer();
        return false;
    }

    /**
     * Support for ftell().
     *
     * @return int
     *   The current offset in bytes from the beginning of file.
     *
     * @see http://php.net/manual/en/streamwrapper.stream-tell.php
     */
    public function stream_tell()
    {
        return $this->position;
    }

    /**
     * Support for fstat().
     *
     * @return array
     *   An array with file status, or FALSE in case of an error - see fstat()
     *   for a description of this array.
     *
     * @see http://php.net/manual/en/streamwrapper.stream-stat.php
     */
    public function stream_stat()
    {
        return $this->_stat();
    }

    /**
     * Support for unlink().
     *
     * @param string $uri
     *   A string containing the uri to the resource to delete.
     *
     * @return bool
     *   TRUE if resource was successfully deleted.
     *
     * @see http://php.net/manual/en/streamwrapper.unlink.php
     */
    public function unlink($uri)
    {
        $this->callConstructor();
        $target = $this->getLocalPath($uri);
        if ($target)
        {
            try
            {
                $response = $this
                    ->gcs
                    ->objects
                    ->delete($this->bucket, $this->folder . '/' . $target);
                if ($response == NULL)
                {
                    return true;
                }
            }
            catch(Exceception $e)
            {
                watchdog_exception('gcs', $e, 'GCSStreamWrapper::unlink');
            }
        }
        return false;
    }

    /**
     * Support for rename().
     *
     * If $to_uri exists, this file will be overwritten. This behavior is
     * identical to the PHP rename() function.
     *
     * @param string $from_uri
     *   The uri to the file to rename.
     * @param string $to_uri
     *   The new uri for file.
     *
     * @return bool
     *   TRUE if file was successfully renamed.
     *
     * @see http://php.net/manual/en/streamwrapper.rename.php
     */
    public function rename($from_uri, $to_uri)
    {
        $this->callConstructor();
        $from = $this->getLocalPath($from_uri);
        $to = $this->getLocalPath($to_uri);

        if ($from && $to)
        {
            try
            {
                $response = $this
                    ->gcs
                    ->objects
                    ->insert($this->bucket, $this->folder . '/' . $to);
                if ($response)
                {
                    $deleted = $this->unlink($from_uri);
                    if ($deleted)
                    {
                        return true;
                    }
                }
            }
            catch(Exception $e)
            {
                watchdog_exception('gcs', $e, 'GCSStreamWrapper::rename');
            }
        }
        return false;
    }

    /**
     * Returns the local writable target of the resource within the stream.
     *
     * This function should be used in place of calls to realpath() or similar
     * functions when attempting to determine the location of a file. While
     * functions like realpath() may return the location of a read-only file, this
     * method may return a URI or path suitable for writing that is completely
     * separate from the URI used for reading.
     *
     * @param string $uri
     *   Optional URI.
     *
     * @return string
     *   Returns a string representing a location suitable for writing of a file,
     *   or FALSE if unable to write to the file such as with read-only streams.
     */
    protected function getTarget($uri = NULL)
    {

        if (!isset($uri))
        {
            $uri = $this->uri;
        }

        list($scheme, $target) = explode('://', $uri, 2);

        // Remove erroneous leading or trailing forward-slashes and backslashes.
        return trim($target, '\/');
    }

    /**
     * Support for mkdir().
     *
     * Google Cloud Storage is not a tree structure, so folder's
     * do not exist.
     *
     * @param string $uri
     *   A string containing the URI to the directory to create.
     * @param int $mode
     *   Permission flags - see mkdir().
     * @param int $options
     *   A bit mask of STREAM_REPORT_ERRORS and STREAM_MKDIR_RECURSIVE.
     *
     * @return bool
     *   TRUE if directory was successfully created.
     *
     * @see http://php.net/manual/en/streamwrapper.mkdir.php
     */
    public function mkdir($uri, $mode, $options)
    {
        // Folders do not exist.
        return true;
    }

    /**
     * Support for rmdir().
     *
     * @param string $uri
     *   A string containing the URI to the directory to delete.
     * @param int $options
     *   A bit mask of STREAM_REPORT_ERRORS.
     *
     * @return bool
     *   TRUE if directory was successfully removed.
     * @see http://php.net/manual/en/streamwrapper.rmdir.php
     */
    public function rmdir($uri, $options)
    {
        return true;
    }

    /**
     * Support for stat().
     *
     * @param string $uri
     *   A string containing the URI to get information about.
     * @param int $flags
     *   A bit mask of STREAM_URL_STAT_LINK and STREAM_URL_STAT_QUIET.
     *
     * @return array
     *   An array with file status, or FALSE in case of an error - see fstat()
     *   for a description of this array.
     *
     * @see http://php.net/manual/en/streamwrapper.url-stat.php
     */
    public function url_stat($uri, $flags)
    {
        $this->callConstructor();
        return $this->_stat($uri);
    }

    /**
     * Gets the name of the directory from a given path.
     *
     * This method is usually accessed through drupal_dirname(), which wraps
     * around the normal PHP dirname() function, which does not support stream
     * wrappers.
     *
     * @param string $uri
     *   An optional URI.
     *
     * @return string
     *   A string containing the directory name, or FALSE if not applicable.
     *
     * @see drupal_dirname()
     */
    public function dirname($uri = NULL)
    {
        list($scheme, $target) = explode('://', $uri, 2);
        $target = $this->getTarget($uri);
        $dirname = dirname($target);

        if ($dirname === '.')
        {
            $dirname = '';
        }

        return $scheme . '://' . $dirname;
    }

    /**
     * Support for opendir().
     *
     * There are no directories in Google Cloud Storage.
     *
     * @param string $uri
     *   A string containing the URI to the directory to open.
     * @param array $options
     *   Unknown (parameter is not documented in PHP Manual).
     *
     * @return bool
     *   TRUE on success.
     *
     * @see http://php.net/manual/en/streamwrapper.dir-opendir.php
     */
    public function dir_opendir($uri, $options)
    {
        return true;
    }

    /**
     * Support for readdir().
     *
     * There are no directories in Google Cloud Storage.
     *
     * @return string|bool
     *   The next filename, or FALSE if there are no more files in the directory.
     *
     * @see http://php.net/manual/en/streamwrapper.dir-readdir.php
     */
    public function dir_readdir()
    {
        return false;
    }

    /**
     * Support for rewinddir().
     *
     * There are no directories in Google Cloud Storage.
     *
     * @return bool
     *   TRUE on success.
     *
     * @see http://php.net/manual/en/streamwrapper.dir-rewinddir.php
     */
    public function dir_rewinddir()
    {
        return true;
    }

    /**
     * Support for closedir().
     *
     * There are no directories in Google Cloud Storage.
     *
     * @return bool
     *   TRUE on success.
     *
     * @see http://php.net/manual/en/streamwrapper.dir-closedir.php
     */
    public function dir_closedir()
    {
        return true;
    }

    /**
     * Return the local filesystem path.
     *
     * @param string $uri
     *   Optional URI, supplied when doing a move or rename.
     */
    protected function getLocalPath($uri = NULL)
    {
        if (!isset($uri))
        {
            $uri = $this->uri;
        }

        $path = str_replace('gs://', '', $uri);
        $path = trim($path, '/');
        return $path;
    }

    /**
     * Flush the object buffers.
     */
    protected function clearBuffer()
    {
        $this->position = 0;
        $this->objectSize = 0;
        $this->buffer = NULL;
        $this->bufferWrite = false;
        $this->bufferLength = 0;
        $this->object = NULL;
    }

    /**
     * Get the Google Cloud Storage connection.
     */
    protected function getGCS()
    {
        if ($this->gcs == NULL)
        {
            if (!libraries_load('google-api-php-client'))
            {
                drupal_set_message(t('Unable to load the Google API PHP library. Please check you have installed the library correctly.') . 'error');
            }
            elseif (!class_exists('Google_Client'))
            {
                watchdog('GCS', 'Google Client class not defined.');
            }
            elseif (empty($this->bucket))
            {
                watchdog('GCS', 'Bucket name not configured.');
            }
            else
            {
                try
                {

                    $http = new GuzzleHttp\Client(['verify' => 'C:\xampp\php\extras\ssl\cacert.pem']);
                    $client = new Google_Client();

                    $client->setApplicationName($this->clientEmail);
                    $client->setDeveloperKey($this->private_key);
                    $cred = $client->setAuthConfig($this->keyFilePath);
                    $client->setScopes(array(
                        'https://www.googleapis.com/auth/devstorage.read_write'
                    ));
                    $service = new Google_Service_Storage($client);

                    if ($client->isAccessTokenExpired())
                    {
                        $client->refreshTokenWithAssertion();
                    }

                    $token = $client->getAccessToken();
                    $client->setAccessToken($token);

                    // Store services for later use.
                    $this->gcs = $service;
                    $this->client = $client;
                    $this->cred = $cred;
                }
                catch(Exception $e)
                {
                    watchdog_exception('gcs', $e, 'GCSStreamWrapper::getGCS');
                }
            }
        }
    }

    /**
     * Creates a signed url for the content.
     *
     * @param string $file_path
     *   The entire filepath to the object.
     *
     * @return null|string
     *   A signed URL or NULL.
     */
    protected function generateSignedURL($file_path)
    {

        $expiry = time() + variable_get('gcs_signed_url_expiry');
        $file_info = explode('/', $file_path);
        $file_name = rawurlencode(array_pop($file_info));
        if ($file_info[0] == "")
        {
            $file_path = implode('/', $file_info) . $file_name;
        }
        else
        {
            $file_path = implode('/', $file_info) . '/' . $file_name;
        }
        $string_policy = "GET\n\n\n" . $expiry . "\n/" . $this->bucket . '/' . $this->folder . '/' . $file_path;
        $url = NULL;
        if ($string_policy)
        {
            $secure = 'http://';
            if ($this->https)
            {
                $secure = 'https://';
            }
            $url = $secure . $this->bucket . '.' . $this->domain . '/' . $this->folder . '/' . $file_path;

        }
        return $url;
    }

    /**
     * Get file status.
     *
     * @param string $uri
     *   The uri to the file.
     *
     * @return array
     *   An array with file status, or FALSE in case of an error - see fstat()
     *   for a description of this array.
     */
    protected function _stat($uri = NULL)
    {
        if ($this->client == NULL)
        {
            return false;
        }
        if (!isset($uri))
        {
            $uri = $this->uri;
        }
        $this->uri = $uri;
        $target = $this->getTarget($uri);
        if ($target)
        {
            try
            {
                $path_info = pathinfo($this->uri);
                if (!isset($path_info['extension']))
                {
                    $mode = 0040000;
                    $size = 1;
                    $mode |= 0777;
                    $stat = $this->createFileMode($mode, $size);
                    return $stat;
                }
                $response = $this
                    ->gcs
                    ->objects
                    ->get($this->bucket, $this->folder . '/' . $target);
                if ($response && isset($response->size) && $response->size != NULL)
                {
                    $mode = 0100000;
                    $size = $response->size;
                    $mode |= 0777;
                    $stat = $this->createFileMode($mode, $size);
                    return $stat;
                }
            }
            catch(Exception $e)
            {
            }
        }
        return false;
    }

    /**
     * Creates the fstat array.
     *
     * @param int $mode
     *   The file mode.
     * @param int $size
     *   The file size.
     *
     * @return array
     *   An array with fstat info populated or empty.
     */
    protected function createFileMode($mode, $size)
    {
        $stat = array();

        $stat[0] = $stat['dev'] = 0;
        $stat[1] = $stat['ino'] = 0;
        $stat[2] = $stat['mode'] = $mode;
        $stat[3] = $stat['nlink'] = 0;
        $stat[4] = $stat['uid'] = 0;
        $stat[5] = $stat['gid'] = 0;
        $stat[6] = $stat['rdev'] = 0;
        $stat[7] = $stat['size'] = 0;
        $stat[8] = $stat['atime'] = 0;
        $stat[9] = $stat['mtime'] = 0;
        $stat[10] = $stat['ctime'] = 0;
        $stat[11] = $stat['blksize'] = 0;
        $stat[12] = $stat['blocks'] = 0;
        $stat[4] = $stat['uid'] = 0;
        $stat[7] = $stat['size'] = $size;
        $stat[8] = $stat['atime'] = 0;
        $stat[9] = $stat['mtime'] = 0;
        $stat[10] = $stat['ctime'] = 0;

        return $stat;
    }

    /**
     * Assert that the constructor has been called, call it if not.
     *
     * Due to PHP bug #40459, the constructor of this class isn't always called
     * for some of the methods. This private method calls the constructor if
     * it hasn't been called before.
     *
     * @see https://bugs.php.net/bug.php?id=40459
     */
    protected function callConstructor()
    {
        if ($this->gcs == NULL)
        {
            $this->__construct();
        }

        if (isset($this->client) && $this
            ->client
            ->isAccessTokenExpired())
        {
            $this
                ->client
                ->refreshTokenWithAssertion($this->cred);
        }
    }
}

